import '../struct/named_struct.dart';
import '../variable/variable.dart';
import '../../grammar/constant.dart';
import '../entity.dart';
import '../function/function.dart';
import '../../value/namespace/namespace.dart';
// import '../../shared/stringify.dart' as util;
import '../../shared/jsonify.dart' as util;
import '../../type/type.dart';
import '../../type/structural.dart';
import '../../error/error.dart';
import '../../interpreter/interpreter.dart';
import '../../declaration/declaration.dart';

/// A prototype based dynamic object.
/// You can define and delete members in runtime.
/// Use prototype to create and extends from other object.
/// Can be named or anonymous.
/// Unlike class, you have to use 'this' to
/// access struct member within its own methods
class HTStruct with HTEntity {
  static var structLiteralIndex = 0;

  final HTInterpreter interpreter;

  final String? id;

  HTStruct? prototype;

  final bool isRootPrototype;

  HTNamedStruct? declaration;

  final Map<String, dynamic> _fields;

  late final HTNamespace namespace;

  final HTNamespace? closure;

  @override
  HTStructuralType get valueType {
    final fieldTypes = <String, HTType>{};
    for (final key in _fields.keys) {
      final value = _fields[key];
      final encap = interpreter.encapsulate(value);
      fieldTypes[key] = encap.valueType?.resolve(namespace) ??
          HTTypeAny(interpreter.lexicon.typeAny);
    }
    return HTStructuralType(namespace, fieldTypes: fieldTypes);
  }

  HTStruct(this.interpreter,
      {String? id,
      this.prototype,
      this.isRootPrototype = false,
      Map<String, dynamic>? fields,
      this.closure})
      : id = id ??
            '${InternalIdentifier.anonymousStruct}${structLiteralIndex++}',
        _fields = fields ?? {} {
    namespace = HTNamespace(
        lexicon: interpreter.lexicon, id: this.id, closure: closure);
    namespace.define(interpreter.lexicon.kThis,
        HTVariable(id: interpreter.lexicon.kThis, value: this));
  }

  Map<String, dynamic> toJson() => util.jsonifyStruct(this);

  // @override
  // String toString() {
  //   if (_fields.isNotEmpty) {
  //     final content =
  //         interpreter.lexicon.stringifyStructMembers(this, from: this);
  //     return '${interpreter.lexicon.codeBlockStart}\n$content${interpreter.lexicon.codeBlockEnd}';
  //   } else {
  //     return '${interpreter.lexicon.codeBlockStart}${interpreter.lexicon.codeBlockEnd}';
  //   }
  // }

  /// Check if this struct has the key in its own _fields
  bool containsKey(String? varName) {
    return _fields.containsKey(varName);
  }

  /// Check if this struct has the key in its own _fields or its prototypes' _fields
  @override
  bool contains(String? varName) {
    if (varName == null) {
      return false;
    }

    if (_fields.containsKey(varName)) {
      return true;
    } else if (prototype != null && prototype!.contains(varName)) {
      return true;
    } else {
      return false;
    }
  }

  void import(HTStruct other, {bool clone = false}) {
    for (final key in other._fields.keys) {
      if (!_fields.keys.contains(key)) {
        define(key, other._fields[key]);
      }
    }
  }

  void define(String id, dynamic value,
      {bool override = false, bool throws = true}) {
    _fields[id] = value;
  }

  void delete(String id) {
    if (_fields.containsKey(id)) {
      _fields.remove(id);
    }
  }

  operator [](dynamic key) {
    return memberGet(key);
  }

  operator []=(dynamic key, dynamic value) {
    memberSet(key, value);
  }

  Iterable<String> get keys => _fields.keys;

  Iterable<dynamic> get values => _fields.values;

  /// The number of key/value pairs in the map.
  int get length => _fields.length;

  /// Whether there is no key/value pair in the map.
  bool get isEmpty => _fields.isEmpty;

  /// Whether there is at least one key/value pair in the map.
  bool get isNotEmpty => _fields.isNotEmpty;

  /// [isSelf] means wether this is called by the struct itself, or a recursive one
  @override
  dynamic memberGet(dynamic varName,
      {String? from, bool isRecursivelyGet = false}) {
    if (varName == null) {
      return null;
    }
    if (varName is! String) {
      varName = varName.toString();
    }
    if (varName == InternalIdentifier.prototype) {
      return prototype;
    }

    dynamic value;
    final getter = '${InternalIdentifier.getter}$varName';
    final constructor = varName != id
        ? '${InternalIdentifier.namedConstructorPrefix}$varName'
        : InternalIdentifier.defaultConstructor;

    if (_fields.containsKey(varName)) {
      if (interpreter.lexicon.isPrivate(varName) &&
          from != null &&
          !from.startsWith(namespace.fullName)) {
        throw HTError.privateMember(varName);
      }
      value = _fields[varName];
    } else if (_fields.containsKey(getter)) {
      if (interpreter.lexicon.isPrivate(varName) &&
          from != null &&
          !from.startsWith(namespace.fullName)) {
        throw HTError.privateMember(varName);
      }
      value = _fields[getter]!;
    } else if (_fields.containsKey(constructor)) {
      if (interpreter.lexicon.isPrivate(varName) &&
          from != null &&
          !from.startsWith(namespace.fullName)) {
        throw HTError.privateMember(varName);
      }
      value = _fields[constructor]!;
    } else if (prototype != null) {
      value = prototype!.memberGet(varName, from: from, isRecursivelyGet: true);
    }

    if (value is HTDeclaration) {
      value.resolve();
    }
    // assign the original struct as instance, not the prototype object
    if (!isRecursivelyGet) {
      if (value is HTFunction) {
        value.namespace = namespace;
        value.instance = this;
        if (value.category == FunctionCategory.getter) {
          value = value.call();
        }
      }
    }
    return value;
  }

  @override
  bool memberSet(dynamic varName, dynamic varValue,
      {String? from, bool defineIfAbsent = true, bool recursive = true}) {
    if (varName == null) {
      throw HTError.nullSubSetKey();
    }
    if (varName is! String) {
      varName = varName.toString();
    }
    if (varName == InternalIdentifier.prototype) {
      if (varValue is! HTStruct) {
        throw HTError.notStruct();
      }
      prototype = varValue;
      return true;
    }

    final setter = '${InternalIdentifier.setter}$varName';
    if (_fields.containsKey(varName)) {
      if (interpreter.lexicon.isPrivate(varName) &&
          from != null &&
          !from.startsWith(namespace.fullName)) {
        throw HTError.privateMember(varName);
      }
      _fields[varName] = varValue;
      return true;
    } else if (_fields.containsKey(setter)) {
      if (interpreter.lexicon.isPrivate(varName) &&
          from != null &&
          !from.startsWith(namespace.fullName)) {
        throw HTError.privateMember(varName);
      }
      HTFunction func = _fields[setter]!;
      func.namespace = namespace;
      func.instance = this;
      func.call(positionalArgs: [varValue]);
      return true;
    } else if (recursive && prototype != null) {
      final success = prototype!
          .memberSet(varName, varValue, from: from, defineIfAbsent: false);
      if (success) {
        return true;
      }
    }
    if (defineIfAbsent) {
      _fields[varName] = varValue;
      return true;
    }
    return false;
  }

  @override
  dynamic subGet(dynamic varName, {String? from}) =>
      memberGet(varName, from: from);

  @override
  void subSet(dynamic varName, dynamic varValue, {String? from}) =>
      memberSet(varName, varValue, from: from);

  // return a deep copy of this struct.
  HTStruct clone() {
    final cloned =
        HTStruct(interpreter, prototype: prototype, closure: closure);
    for (final key in _fields.keys) {
      final value = _fields[key];
      final copiedValue = interpreter.toStructValue(value);
      cloned.define(key, copiedValue);
    }
    return cloned;
  }

  // deep copy another struct then assign to this one.
  void assign(HTStruct other) {
    for (final key in other._fields.keys) {
      if (key.startsWith(interpreter.lexicon.internalPrefix)) continue;
      final value = other._fields[key];
      final copiedValue = interpreter.toStructValue(value);
      define(key, copiedValue);
    }
  }
}
